// ***********************************************************************
// syssoft_linux.cpp
// ***********************************************************************

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/mman.h>

#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>
#include <semaphore.h>
#include <stdio.h>

#include <string>

// ***********************************************************************
// System
// ***********************************************************************

class System {
public:
	static void perror ( const char *comment ) {
		::perror(comment);
		exit(0);
	}

	static void fatal ( const char *comment ) {
		fprintf(stderr,"%s\n",comment);
		exit(0);
	}
};

// ***********************************************************************
// Moment
// ***********************************************************************

class Moment {
public:
	void now () {
		int ret = gettimeofday(&tv,NULL);
		if (ret == -1) System::perror("Moment::now: gettimeofday()");
	}

	double operator - ( Moment &m ) {
		double result = ((double)  tv.tv_sec - m.tv.tv_sec) * 1e6;
		result += ((double) tv.tv_usec - m.tv.tv_usec);
		return result/1000.0;
	}
private:
	struct timeval tv;
};

// ***********************************************************************
// Thread
// ***********************************************************************

typedef void *(*Thread_Main_T)(void *);

class Thread {
public:
	Thread ( Thread_Main_T p_main ) : thread_main(p_main) {};
	~Thread () {};

	void run ( void *arg ) {
		int ret = pthread_create(&pid,NULL,thread_main,arg);
		if (ret != 0) System::perror("Thread::run: pthread_create()");
	}

	void join () {
		int ret = pthread_join(pid,NULL);
		if (ret != 0) System::perror("Thread::join: pthread_join()");
	}

	pthread_t getpid () const { return pid; };
private:
	pthread_t pid;
	Thread_Main_T thread_main;
};

// ***********************************************************************
// Process
// ***********************************************************************

class Process {
public:
	Process ( char *executable, char **args, char **env );
	Process ( Thread_Main_T arg_main, void *arg_arg );
	~Process ();

	int wait ();
private:
	pid_t pid;
	bool terminated;
};

Process::Process ( char *executable, char **args, char **env ) : terminated(false) {
	pid = fork();
	if (pid == -1) System::perror("Process::Process: fork()");
	if (pid == 0) {
		int r = execve(executable,args,env);
		if (r == -1) System::perror("Process::Process: execve()");
	}
	// parent will leave constructor immediately
}

Process::Process ( Thread_Main_T arg_main, void *arg_arg ) {
	pid = fork();
	if (pid == -1) System::perror("Process::Process: fork()");
	if (pid == 0) {
		(void) (arg_main)(arg_arg);
	}
	// parent will leave constructor immediately
}

Process::~Process () {
	if (terminated)
		return;
	pid_t r = waitpid(pid,NULL,WNOHANG);
	if (r == -1) System::perror("Process::~Process: waitpid()");
	if (r == 0) {
		// Not terminated yet, so kill it explicitly
		int rv = kill(pid,SIGKILL);
		if (rv == -1) System::perror("Process::~Process: kill()");
	}
}

int Process::wait () {
	int retval;

	pid_t r = waitpid(pid,&retval,0);
	if (r == -1) System::perror("Process::wait: waitpid()");
	terminated = true;
	if (WIFEXITED(retval))
		return WEXITSTATUS(retval);
	else
		return 0;
}

// ***********************************************************************
// Semaphore
// ***********************************************************************

class Semaphore {
public:
	Semaphore ( int p_initial ) {
		int ret = sem_init(&sema,0,p_initial);
		if (ret != 0) System::perror("Semaphore::Semaphore: sem_init()");
	}
	~Semaphore () {
		int ret = sem_destroy(&sema);
		if (ret != 0) System::perror("Semaphore::~Semaphore: sem_destroy()");
	}

	void P () {
		sem_wait(&sema);
	}
	void V () {
		int ret = sem_post(&sema);
		if (ret != 0) System::perror("Semaphore::V: sem_post()");
	}

	int get_value () {
		int v;
		sem_getvalue(&sema,&v);
		return v;
	}
private:
	sem_t sema;
};

// ***********************************************************************
// SharedMemory
// ***********************************************************************

class SharedMemory {
public:
	SharedMemory ( const std::string &p_filename, int p_size )
	: filename(p_filename), len(p_size) {
		fd = shm_open(filename.c_str(),O_RDWR|O_CREAT,0x1ff);
		if (fd == -1) System::perror("SharedMemory::SharedMemory: shm_open()");
		int ret = ftruncate(fd,len);
		if (ret == -1) System::perror("SharedMemory::SharedMemory: ftruncate()");
		addr = mmap(0,len,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
		if (addr == MAP_FAILED) System::perror("SharedMemory::SharedMemory: mmap()");	
	}

	SharedMemory ( const std::string &p_filename )
	: filename(p_filename) {
		struct stat buf;
		std::string path = "/dev/shm/" + filename;
		int ret = stat(path.c_str(),&buf);
		if (ret == -1) System::perror("SharedMemory::SharedMemory: stat()");
		len = buf.st_size;
		fd = shm_open(filename.c_str(),O_RDWR,0);
		if (fd == -1) System::perror("SharedMemory::SharedMemory: shm_open()");
		addr = mmap(0,len,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
		if (addr == MAP_FAILED) System::perror("SharedMemory::SharedMemory: mmap()");	
	}

	~SharedMemory () {
		int ret = munmap(addr,len);
		if (ret != 0) System::perror("SharedMemory::~SharedMemory: munmap()");
		// do not delete by calling shm_unlink(filename.c_str());
	}

	char *base () const {
		return reinterpret_cast<char *>(addr);
	}

	int size () const {
		return len;
	}

private:
	std::string filename;
	int fd;
	void *addr;
	int len;
};


// ***********************************************************************
// MappedFile
// ***********************************************************************

class MappedFile {
public:
	MappedFile ( std::string filename ) {
		struct stat buf;
		int ret = stat(filename.c_str(),&buf);
		if (ret == -1) System::perror("MappedFile::MappedFile: stat()");
		len = buf.st_size;
		fd = open(filename.c_str(),O_RDWR);
		if (fd == -1) System::perror("MappedFile::MappedFile: open()");
		addr = mmap(0,len,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
		if (addr == MAP_FAILED) System::perror("MappedFile::MappedFile: mmap()");		
	}
	~MappedFile () {
		int ret = munmap(addr,len);
		if (ret != 0) System::perror("MappedFile::~MappedFile: munmap()");
	}

	char *base () const {
		return reinterpret_cast<char *>(addr);
	}

	int size () const {
		return len;
	}
private:
	int fd;
	void *addr;
	int len;
};






